%{
/*
       Jacek Pospychala jacekp@siat.pl grupa:poniedzialki, 15:10  
*/
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
  int yylex(void);
  void yyerror(const char *,...);
  void add_func(char*,int,int);
  void func_exists(char*,int,int);
  void add_var(char*); /* adds var in current namespace */
  void del_var(char*); /* dels var in current namespace */
  void set_var(char*,int);
  int var_val(char*);
  void print_var(char*);
  void start_varscope();
  void end_varscope();
  int yyparse(void);
  
  extern int lineno;
  
  struct {
    char *name;
    int val;
  } vars[100] = {0};
  int varscount=0; 
  
  int varscopes[10] = {0}; 
  int currentvs=0;
  
  struct {
    char *funcname;
    int  argcount;
    int  returntype;
  } funcs[100] = {0};
  int funcscount=0;
  int declvartype;
 
%}
%union
{
  char *text;
  int inum;
}
%token <inum> VOID INT CHAR LONG
%token <inum> STATIC REGISTER CONST
%token <text> ID STRING
%token <inum> NUM 
%token <inum> IF ELSE
%token <inum> '*'
%token <inum> DIR_PRINT RETURN
%type <inum> declaration_specifiers type_specifier storage_class_specifier declaration start parameter_list declaration_list
%type <inum> parameter_declaration external_declaration function_definition compiler_directive assignment
%type <text> declarator direct_declarator init_declarator init_declarator_list
%type <inum> assignment_value mul_value simple_value if_statement function_parameter_list
%%
start       : start external_declaration
            | external_declaration
            ;
external_declaration : declaration
                     | function_definition
                     ;

function_definition : declaration_specifiers declarator compound_statement
                    ;

compound_statement : '{' '}'
                   | '{' { start_varscope(); } declaration_list '}' { end_varscope(); }
                   ;

declaration_list : declaration
                 | declaration declaration_list
                 ;

declaration : declaration_specifiers init_declarator_list ';' { declvartype=$1; }
            | compiler_directive ';'
	    | assignment ';'
	    | assignment_value ';'
            | if_statement
	    | RETURN NUM ';'
            ;

compiler_directive : DIR_PRINT ID { print_var($2); }
                   ;

assignment : ID '=' assignment_value { set_var($1, $3); }
           ;

if_statement : IF '(' assignment_value ')' compound_statement { $$=IF; }
             | if_statement ELSE compound_statement {$$=ELSE;if ($1!=IF)yyerror("Error on line %d, problem wiszacego else.\n",lineno);}
             ;

declaration_specifiers : type_specifier                                
                       | type_specifier declaration_specifiers            
                       | storage_class_specifier                        
                       | storage_class_specifier declaration_specifiers { if ($1==$2) yyerror("Duplicated qualifier on line %d\n", lineno); }
                       ;

init_declarator_list : init_declarator                          
                     | init_declarator ',' init_declarator_list
                     ;

init_declarator : declarator
                | declarator '=' assignment_value { set_var($1, $3); }
                ;

declarator : '*' direct_declarator { $$=$2; }
           | direct_declarator
           ;

direct_declarator : ID                                       { add_var($1); }
                  | direct_declarator '(' parameter_list ')' { del_var($1); add_func($1, declvartype, $3); $$=$1; }
                  | direct_declarator '(' ')'                { del_var($1); add_func($1, declvartype, 0); $$=$1; }
                  ;

parameter_list : parameter_declaration { $$=1; }
               | parameter_declaration ',' parameter_list { $$=$3+1; }
               ;

parameter_declaration : declaration_specifiers declarator
                      ;

assignment_value : assignment_value '+' mul_value { $$=$1+$3; }
                 | assignment_value '-' mul_value { $$=$1-$3; }
                 | mul_value
                 ;

mul_value : mul_value '*' simple_value { $$=$1*$3; }
          | mul_value '/' simple_value { $$=$1/$3; }
          | simple_value
          ;
 
simple_value : '(' assignment_value ')' { $$=$2; }
             | NUM
             | STRING                   { $$=0; }
             | ID                       { $$=var_val($1); }
             | ID '(' function_parameter_list ')' { func_exists($1, NUM, $3); $$=0; }
             | ID '(' ')'               { func_exists($1, NUM, 0); $$=0; }
             ;

function_parameter_list : function_parameter_list ',' simple_value  {$$=$1+1;}
                        | simple_value                              {$$=1;}
                        ;

type_specifier : VOID
               | INT 
               | LONG
               | CHAR
               | { yyerror("Undeclared identifier on line %d.\n", lineno); } error
               ;
storage_class_specifier : STATIC  
                        | REGISTER
                        | CONST   
                        ;
%%
void add_func(char *funcname, int returntype, int argcount) {
  int i=0;
  while ((i<funcscount) && (
			    (strcmp(funcname, funcs[i].funcname)!=0) ||
			    (returntype!=funcs[i].returntype) || 
			    (argcount!=funcs[i].argcount)
			    ))
        i++;

  if (i<funcscount)
    yyerror("Error on line %d, function '%s' is already defined.\n", lineno, funcname);
  else 
    {
      // printf("added function '%s' with %d args returns %d.\n", funcname, argcount, returntype);
      funcs[funcscount].funcname = strdup(funcname);
      funcs[funcscount].returntype = returntype;
      funcs[funcscount].argcount = argcount;
      funcscount++;
    }
}

void func_exists(char *funcname,int returntype,int argcount)
{
  int i=0;
  // printf("searching function '%s' with %d args, which returns %d.\n", funcname, argcount, returntype);
  while ((i<funcscount) && (strcmp(funcname, funcs[i].funcname)!=0))
        i++;

  if (i>=funcscount)
    yyerror("Error on line %d, function '%s' is undefined.\n", lineno, funcname);
}

void add_var(char *varname) {
  int i=varscopes[currentvs];
  while ((i<varscount) && (strcmp(vars[i].name, varname)!=0)) i++;
  if (i<varscount)
    yyerror("Error on line %d, identifier '%s' is already defined.\n", lineno, varname);
  else
    {
      //printf("var %s added at %d (varscope:%d)\n", varname, varscount, currentvs);
      vars[varscount].name = strdup(varname);
      varscount++;
    }
}

void del_var(char *varname) {
  int i=varscopes[currentvs];
  while ((i<varscount) && (strcmp(vars[i].name, varname)!=0)) i++;
  if (i<varscount)
    {
      //printf("var %s deleted at %d (varscope:%d)\n", vars[i], i, currentvs);
      varscount--;
      vars[i].name = vars[varscount].name;
    }
} 

void set_var(char *varname,int value)
{
  int i=varscount-1;
  while ((i>=0) && (strcmp(vars[i].name, varname) !=0)) i--;
  if (i>=0)
    vars[i].val = value;
  else
    yyerror("Error on line %d, identifier '%s' is undefined.\n", lineno, varname);
}

int var_val(char *varname)
{
 int i=varscount-1;
 while ((i>=0) && (strcmp(vars[i].name, varname) !=0)) i--;
 if (i>=0)
   return vars[i].val;
 else
   yyerror("Error on line %d, identifier '%s' is undefined.\n", lineno, varname);
}

void print_var(char *varname)
{
  printf("%d\n", var_val(varname));
}

void start_varscope()
{
  currentvs++;
  varscopes[currentvs] = varscount;
  //printf("start varscope %d with %d known vars.\n", currentvs, varscount);
}

void end_varscope()
{
  varscount = varscopes[currentvs];
  currentvs--;
  //printf("back to varscope %d with %d vars.\n", currentvs, varscount);
}

void yyerror(const char *fmt,...) {
  if (strcmp(fmt, "parse error")==0)
    printf("Syntax error on line %d.\n", lineno);
  else {
    va_list ap;
    va_start(ap, fmt);
    vprintf(fmt, ap);
    va_end(ap);
  }
  exit(1); 
}
int main() { if (yyparse()==0) printf("OK\n"); }
