%{
  #include <stdio.h>
  #include <stdarg.h>
  #include <stdlib.h>
  void yyerror(char *fmt,...);
  int main(void);
  void add_var(char*,int);
  void ns_start(void);
  void ns_end(void);
  int lineno=0;

  int ns[100]; // namespaces stack
  int nsp=0;   // namespaces stack pointer

  struct {
    char* name;
    int type; // 0-variable, 1-class, 2-function
    int varsc;
    int funcsc;
    int classc;
  } vstack[100]; // variables stack
  int vstackp = 0;  // variables stack pointer
  int lastclass = -1;

%}
%union
{
  int inum;
  char *text;
}
%token <text> ID
%token <inum> NUM CLASS INT CHAR
%type <inum> class_unit exec_unit translation_unit
%type <inum> var_type c_func c_arg class_element class_elements_list
%type <inum> assignment decl_list decl farg_list farg var_type
%%
translation_unit 
: exec_unit translation_unit
| exec_unit 
;

exec_unit
: class_unit
;

class_unit
: CLASS ID {add_var($2, 1); } compound_statement
;

class_elements_list
: class_element class_elements_list
| class_element

;

class_element
: c_arg
| c_func
| class_unit
| compound_statement {$$=0;}
;

c_arg
: var_type decl_list ';'
;

c_func
: var_type ID '(' farg_list ')' { add_var($2, 2); } compound_statement 
| var_type ID '(' ')' { add_var($2, 2); } compound_statement 
;

var_type
: INT
| CHAR
;

decl_list
: decl ',' decl_list
| decl
;

decl
: ID '=' assignment { $$=0; add_var($1, 0); }
| ID { $$=0; add_var($1, 0); }
;

farg_list
: farg ',' farg_list
| farg
;

farg
: var_type ID
;

compound_statement
: '{' {  ns_start();}  class_elements_list '}'     {  ns_end(); }
| '{' '}'
| '{' { yyerror("Closing bracket '}' expected\n"); }
;

assignment
: NUM
;

%%
void yyerror(char *fmt, ...)
{
  va_list ap;
  printf("Syntax error on line %d: ", lineno);
  va_start(ap, fmt);
  vprintf(fmt, ap);
  exit(1);
}
int main(void)
{
  if (yyparse()==0) 
    {
      int i;
      printf("declared %d vars:\n", vstackp);
      for (i=0; i<vstackp; i++)
	{
	  switch (vstack[i].type)
	    {
	    case 0: printf("variable '%s'\n", vstack[i].name); break;
	    case 1: printf("class '%s': %d fields, %d methods, %d subclasses\n", vstack[i].name, vstack[i].varsc, vstack[i].funcsc, vstack[i].classc); break;
	    case 2: printf("function '%s'\n", vstack[i].name); break;
	    }
	}
      printf("OK\n");
    }
}


void add_var(char*name, int type)
{
  int i;
  for (i=ns[nsp]; i<vstackp; i++)
    if (strcmp(name, vstack[i].name)==0) yyerror("Variable '%s' already exists\n", name);

  vstack[vstackp].name = (char*)strdup(name);
  vstack[vstackp].type = type;
  vstackp++;

  if (lastclass>=0)
    {
      switch (type)
	{
	case 0: vstack[lastclass].varsc++; break;
	case 1: vstack[lastclass].classc++; break;
	case 2: vstack[lastclass].funcsc++; break;
	}
    }
  if (type==1) // class
    {
      lastclass = vstackp-1;
    }
}

void ns_start()
{
  //  printf("ns_start\n");
  nsp++;
  ns[nsp]=vstackp;
}

void ns_end()
{
  //printf("ns_end\n");
  vstackp=ns[nsp];
  nsp--;
}
